@@ -85,22 +85,37 @@ class AnthySetup(object):
         builder.add_from_file(builder_file)
 
         toplevel = builder.get_object('main')
+        parent_xid = 0
+        parent_wmname = None
+        parent_wmclass = None
 
         try:
-            gnome_control_center_xid = int(environ['GNOME_CONTROL_CENTER_XID'])
+            parent_xid = int(environ['IBUS_SETUP_XID'])
+            if parent_xid != 0:
+                parent_wmname = 'ibus-setup'
+                parent_wmclass = 'Ibus-setup'
         except:
-            gnome_control_center_xid = 0
+            pass
 
-        if gnome_control_center_xid != 0:
+        try:
+            if parent_xid == 0:
+                parent_xid = int(environ['GNOME_CONTROL_CENTER_XID'])
+                if parent_xid != 0:
+                    parent_wmname = 'gnome-conrol-center'
+                    parent_wmclass = 'Gnome-conrol-center'
+        except:
+            pass
+
+        if parent_xid != 0:
             def set_transient(obj, pspec):
                 window = toplevel.get_window()
                 if window == None:
                     return
                 parent_window = GdkX11.X11Window.foreign_new_for_display(Gdk.Display.get_default(),
-                                                                         gnome_control_center_xid)
+                                                                         parent_xid)
                 if parent_window != None:
                     window.set_transient_for(parent_window)
-            toplevel.set_wmclass('gnome-control-center', 'Gnome-control-center')
+            toplevel.set_wmclass(parent_wmname, parent_wmclass)
             toplevel.set_modal(True)
             toplevel.set_type_hint(Gdk.WindowTypeHint.DIALOG)
             toplevel.connect('notify::window', set_transient)
@@ -328,11 +343,11 @@ class AnthySetup(object):
         return (section, key)
 
     def __run_message_dialog(self, message, type=Gtk.MessageType.INFO):
-        dlg = Gtk.MessageDialog(parent=self.__builder.get_object('main'),
-                                flags=Gtk.DialogFlags.MODAL,
-                                message_type=type,
-                                buttons=Gtk.ButtonsType.OK,
-                                message_format=message)
+        dlg = Gtk.MessageDialog(
+                transient_for=self.__builder.get_object('main'),
+                message_type=type,
+                buttons=Gtk.ButtonsType.OK,
+                text=message)
         dlg.run()
         dlg.destroy()
 
@@ -425,6 +440,7 @@ class AnthySetup(object):
         for key, value in sorted(list(rule.items()), \
             key = functools.cmp_to_key(self.__japanese_tuple_sort)):
             ls.append(['romaji', key, value])
+        tv.set_model(None)
         tv.append_column(Gtk.TreeViewColumn(_(_("Input Chars")),
                                             Gtk.CellRendererText(), text=1))
         tv.append_column(Gtk.TreeViewColumn(_(_("Output Chars")),
@@ -460,6 +476,7 @@ class AnthySetup(object):
         for key, value in sorted(list(rule.items()), \
             key = functools.cmp_to_key(self.__japanese_tuple_sort)):
             ls.append(['kana', key, value])
+        tv.set_model(None)
         tv.append_column(Gtk.TreeViewColumn(_(_("Input Chars")),
                                             Gtk.CellRendererText(), text=1))
         tv.append_column(Gtk.TreeViewColumn(_(_("Output Chars")),
@@ -507,6 +524,7 @@ class AnthySetup(object):
         for key, value in sorted(list(rule.items()), \
             key = functools.cmp_to_key(self.__japanese_thumb_sort)):
             ls.append(['thumb', key, value[0], value[2], value[1]])
+        tv.set_model(None)
         tv.append_column(Gtk.TreeViewColumn(_(_("Input")),
                                             Gtk.CellRendererText(), text=1))
         tv.append_column(Gtk.TreeViewColumn(_(_("Single")),
@@ -587,9 +605,10 @@ class AnthySetup(object):
         combobox.clear()
         combobox.disconnect_by_func(self.on_cb_custom_key_table_changed)
 
-    def __run_dialog_custom_key_table(self, mode):
+    def __run_dialog_custom_key_table(self, widget, mode):
         prefs = self.prefs
         dlg = self.__builder.get_object('dialog_custom_key_table')
+        dlg.set_transient_for(widget.get_toplevel())
         label = self.__builder.get_object('label_custom_key_table')
         label_output = self.__builder.get_object('label_output_custom_key')
         list_labels = []
@@ -1102,6 +1121,72 @@ class AnthySetup(object):
         if self.__keymap != None:
             self.__save_keymap()
 
+    def __search_and_mark(self, buffer, text, start, end, onetime, forward):
+        if forward:
+            match = start.forward_search(text, 0, end)
+        else:
+            match = start.backward_search(text, 0, end)
+        if match == None:
+            return False
+        match_start, match_end = match
+        if onetime:
+            buffer.place_cursor(match_start)
+            buffer.select_range(match_start, match_end)
+            return True
+        buffer.apply_tag(buffer.tag_found, match_start, match_end)
+        self.__search_and_mark(buffer, text, match_end, end, onetime, forward)
+        return True
+
+    def __filter_search(self, entry, onetime, forward):
+        text = entry.get_text()
+        self.__filter_timeout_id = 0
+
+        text_view = entry.text_view
+        buffer = text_view.get_buffer()
+        start = buffer.get_start_iter()
+        if onetime:
+            bounds = buffer.get_selection_bounds()
+            if len(bounds) != 0:
+                start, end = bounds
+                if forward:
+                    start = end
+        end = buffer.get_end_iter()
+        if not forward:
+            end = buffer.get_start_iter()
+        if not onetime:
+            buffer.remove_all_tags(start, end)
+        if text == '':
+            return
+        found = self.__search_and_mark(buffer, text, start, end, onetime, forward)
+        if not found and onetime and forward:
+            end = start
+            start = buffer.get_start_iter()
+            self.__search_and_mark(buffer, text, start, end, onetime, forward)
+
+    def __do_filter(self, entry):
+        self.__filter_search(entry, False, True)
+        return False
+
+    def __filter_changed(self, entry):
+        if self.__filter_timeout_id != 0:
+            return
+        self.__filter_timeout_id = GLib.timeout_add(150,
+                                                    self.__do_filter,
+                                                    entry)
+
+    def __filter_key_release_event(self, entry, event):
+        pressed, keyval = event.get_keyval()
+        if keyval == IBus.KEY_Return:
+            forward = True
+            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
+                forward = False
+            self.__filter_search(entry, True, forward)
+            text_view = entry.text_view
+            buffer = text_view.get_buffer()
+            text_view.scroll_to_mark(buffer.get_insert(),
+                                     0.25, False, 0.0, 0.0)
+        return False
+
     def on_selection_changed(self, widget, id):
         set_sensitive = lambda a, b: self.__builder.get_object(a).set_sensitive(b)
         flg = True if widget.get_selected()[1] else False
@@ -1128,6 +1213,7 @@ class AnthySetup(object):
             Gtk.main_quit()
             return True
         dlg = self.__builder.get_object('quit_check')
+        dlg.set_transient_for(widget.get_toplevel())
         dlg.set_markup('<big><b>%s</b></big>' % _("Confirmation"))
         dlg.format_secondary_text(
                 _("You are about to close the setup dialog, is that OK?"))
@@ -1144,6 +1230,7 @@ class AnthySetup(object):
             Gtk.main_quit()
             return True
         dlg = self.__builder.get_object('quit_check_without_save')
+        dlg.set_transient_for(widget.get_toplevel())
         dlg.set_markup('<big><b>%s</b></big>' % _("Notice!"))
         dlg.format_secondary_text(
                 _("You are about to close the setup dialog without saving your changes, is that OK?"))
@@ -1219,6 +1306,7 @@ class AnthySetup(object):
         for w in ['es:checkbutton_ctrl', 'es:checkbutton_alt', 'es:checkbutton_shift']:
             self.__builder.get_object(w).set_active(False)
         dlg = self.__builder.get_object('edit_shortcut')
+        dlg.set_transient_for(widget.get_toplevel())
         id = dlg.run()
         dlg.hide()
         if id == Gtk.ResponseType.OK:
@@ -1239,13 +1327,13 @@ class AnthySetup(object):
             self.__builder.get_object('btn_apply').set_sensitive(True)
 
     def on_btn_romaji_custom_table_clicked(self, widget):
-        self.__run_dialog_custom_key_table('romaji')
+        self.__run_dialog_custom_key_table(widget, 'romaji')
 
     def on_btn_kana_custom_table_clicked(self, widget):
-        self.__run_dialog_custom_key_table('kana')
+        self.__run_dialog_custom_key_table(widget, 'kana')
 
     def on_btn_thumb_custom_table_clicked(self, widget):
-        self.__run_dialog_custom_key_table('thumb')
+        self.__run_dialog_custom_key_table(widget, 'thumb')
 
     def on_btn_add_custom_key(self, widget, user_data):
         prefs = self.prefs
@@ -1374,6 +1462,7 @@ class AnthySetup(object):
         for w in ['es:checkbutton_ctrl', 'es:checkbutton_alt', 'es:checkbutton_shift']:
             self.__builder.get_object(w).set_active(False)
         dlg = self.__builder.get_object('edit_shortcut')
+        dlg.set_transient_for(widget.get_toplevel())
         id = dlg.run()
         dlg.hide()
         self.__builder.get_object('es:button_add').show()
@@ -1408,16 +1497,17 @@ class AnthySetup(object):
 
         if Gtk.Buildable.get_name(widget) == 'dict:btn_add':
             dlg = Gtk.FileChooserDialog(title=_("Open Dictionary File"),
-                                        parent=self.__builder.get_object('main'),
-                                        action=Gtk.FileChooserAction.OPEN,
-                                        buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
-                                                 Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
+                                        transient_for=widget.get_toplevel(),
+                                        action=Gtk.FileChooserAction.OPEN)
+            buttons=(_("_Cancel"), Gtk.ResponseType.CANCEL,
+                     _("_Open"), Gtk.ResponseType.OK)
+            dlg.add_buttons(*buttons)
         if Gtk.Buildable.get_name(widget) == 'dict:btn_edit':
             dlg = Gtk.Dialog(title=_("Edit Dictionary File"),
-                             parent=self.__builder.get_object('main'),
-                             flags=Gtk.DialogFlags.MODAL,
-                             buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
-                                      Gtk.STOCK_OK, Gtk.ResponseType.OK))
+                             transient_for=widget.get_toplevel())
+            buttons=(_("_Cancel"), Gtk.ResponseType.CANCEL,
+                     _("_OK"), Gtk.ResponseType.OK)
+            dlg.add_buttons(*buttons)
 
         vbox = self.__builder.get_object('dict:add_extra_vbox')
         if Gtk.Buildable.get_name(widget) == 'dict:btn_add':
@@ -1510,11 +1600,12 @@ class AnthySetup(object):
             lines = str(lines, encoding)
 
         dlg = Gtk.Dialog(title=_("View Dictionary File"),
-                         parent=self.__builder.get_object('main'),
-                         flags=Gtk.DialogFlags.MODAL,
-                         buttons=(Gtk.STOCK_OK, Gtk.ResponseType.OK))
+                         transient_for=widget.get_toplevel())
+        buttons=(_("_OK"), Gtk.ResponseType.OK)
+        dlg.add_buttons(*buttons)
         buffer = Gtk.TextBuffer()
         buffer.set_text (lines)
+        buffer.tag_found = buffer.create_tag('found', background = 'yellow')
         text_view = Gtk.TextView.new_with_buffer(buffer)
         text_view.set_editable(False)
         sw = Gtk.ScrolledWindow()
@@ -1524,6 +1615,24 @@ class AnthySetup(object):
         parent_vbox.add(sw)
         sw.show_all()
         dlg.set_default_size(500, 500)
+        self.__filter_timeout_id = 0
+
+        if hasattr(Gtk, 'SearchEntry') and \
+           hasattr(Gtk, 'SearchEntryClass') and \
+           hasattr(Gtk.SearchEntryClass, 'search_changed'):
+            filter_entry = Gtk.SearchEntry(hexpand = True,
+                                           margin_left = 6,
+                                           margin_right = 6,
+                                           margin_top = 6,
+                                           margin_bottom = 6)
+            filter_entry.text_view = text_view
+            filter_entry.connect('search-changed', self.__filter_changed)
+            filter_entry.connect('key-release-event',
+                                 self.__filter_key_release_event)
+            parent_vbox.add(filter_entry)
+            filter_entry.show_all()
+
+        sw.show_all()
         dlg.run()
         dlg.destroy()
 
@@ -1660,6 +1769,7 @@ class AnthySetup(object):
 
     def on_es_button_run_input_clicked(self, widget):
         dlg = self.__builder.get_object('key_input_dialog')
+        dlg.set_transient_for(widget.get_toplevel())
         dlg.set_markup('<big><b>%s</b></big>' % _("Please press a key (or a key combination)"))
         dlg.format_secondary_text(_("The dialog will be closed when the key is released"))
         id = dlg.run()
@@ -1680,6 +1790,7 @@ class AnthySetup(object):
         s = self.__builder.get_object('es:entry').get_text()
         if not s or not IBus.keyval_from_name(s):
             dlg = self.__builder.get_object('invalid_keysym')
+            dlg.set_transient_for(widget.get_toplevel())
             dlg.set_markup('<big><b>%s</b></big>' % _("Invalid keysym"))
             dlg.format_secondary_text(_("This keysym is not valid"))
             dlg.run()
@@ -1700,6 +1811,7 @@ class AnthySetup(object):
         s = self.__builder.get_object('es:entry').get_text()
         if not s or not IBus.keyval_from_name(s):
             dlg = self.__builder.get_object('invalid_keysym')
+            dlg.set_transient_for(widget.get_toplevel())
             dlg.set_markup('<big><b>%s</b></big>' % _("Invalid keysym"))
             dlg.format_secondary_text(_("This keysym is not valid"))
             dlg.run()
